@@ -16,12 +16,6 @@ by an adjacency Mapping.
 
 """
 
-from meshed.itools import (
-    edges,
-    nodes,
-)
 
-from meshed.dag import (
-    DAG,
-    FuncNode,
-)
+from meshed.dag import DAG
+from meshed.base import FuncNode
