@@ -2,12 +2,12 @@
 adjacency Mapping representation.
 
 """
-from typing import Any, Mapping, Sized, MutableMapping, Iterable
+from typing import Any, Mapping, Sized, MutableMapping, Iterable, Callable, TypeVar
 from itertools import product, chain
 from functools import wraps
+from collections import defaultdict
 
 from i2.signatures import Sig
-from meshed.makers import edge_reversed_graph
 
 
 def _handle_exclude_nodes(func):
@@ -15,22 +15,18 @@ def _handle_exclude_nodes(func):
 
     @wraps(func)
     def _func(*args, **kwargs):
-        kwargs = sig.kwargs_from_args_and_kwargs(
-            args, kwargs, apply_defaults=True
-        )
+        kwargs = sig.kwargs_from_args_and_kwargs(args, kwargs, apply_defaults=True)
         try:
-            _exclude_nodes = kwargs["_exclude_nodes"]
+            _exclude_nodes = kwargs['_exclude_nodes']
         except KeyError:
-            raise RuntimeError(
-                f"{func} doesn't have a _exclude_nodes argument"
-            )
+            raise RuntimeError(f"{func} doesn't have a _exclude_nodes argument")
 
         if _exclude_nodes is None:
             _exclude_nodes = set()
         elif not isinstance(_exclude_nodes, set):
             _exclude_nodes = set(_exclude_nodes)
 
-        kwargs["_exclude_nodes"] = _exclude_nodes
+        kwargs['_exclude_nodes'] = _exclude_nodes
         args, kwargs = sig.args_and_kwargs_from_kwargs(kwargs)
         return func(*args, **kwargs)
 
@@ -208,7 +204,7 @@ def ancestors(g: Mapping, source: Iterable, _exclude_nodes=None):
     >>> g = {
     ...     0: [1, 2],
     ...     1: [2, 3, 4],
-    ...     2: [1, 4],
+    ...     2: [4],
     ...     3: [4]
     ... }
     >>> ancestors(g, [2, 3])
@@ -218,12 +214,13 @@ def ancestors(g: Mapping, source: Iterable, _exclude_nodes=None):
     """
     assert isinstance(source, Iterable)
     source = set(source) - _exclude_nodes
-    if source:
-        _parents = set(parents(g, source)) - _exclude_nodes
-        _ancestors_of_parent = ancestors(g, _parents, _exclude_nodes)
-        return _parents | _ancestors_of_parent
-    else:
+    _parents = (set(parents(g, source)) - source) - _exclude_nodes
+    if not _parents:
         return set()
+    else:
+        _ancestors_of_parent = ancestors(g, _parents, _exclude_nodes)
+
+        return _parents | _ancestors_of_parent
 
 
 def descendants(g: Mapping, source: Iterable, _exclude_nodes=None):
@@ -232,11 +229,11 @@ def descendants(g: Mapping, source: Iterable, _exclude_nodes=None):
     >>> g = {
     ...     0: [1, 2],
     ...     1: [2, 3, 4],
-    ...     2: [1, 4],
+    ...     2: [4],
     ...     3: [4]
     ... }
     >>> descendants(g, [2, 3])
-    {1, 4}
+    {4}
     >>> descendants(g, [4])
     set()
     """
@@ -397,3 +394,33 @@ def topological_sort(g: Mapping):
             _topological_sort_helper(g, i, visited, stack)
 
     return stack
+
+
+from typing import TypeVar
+
+T = TypeVar('T')
+
+
+def edge_reversed_graph(
+    g: Mapping[T, Iterable[T]],
+    dst_nodes_factory: Callable[[], Iterable[T]] = list,
+    dst_nodes_append: Callable[[Iterable[T], T], None] = list.append,
+) -> Mapping[T, Iterable[T]]:
+    """
+    >>> g = dict(a='c', b='cd', c='abd', e='')
+    >>> assert edge_reversed_graph(g) == {'c': ['a', 'b'], 'd': ['b', 'c'], 'a': ['c'], 'b': ['c'], 'e': []}
+    >>> reverse_g_with_sets = edge_reversed_graph(g, set, set.add)
+    >>> assert reverse_g_with_sets == {'c': {'a', 'b'}, 'd': {'b', 'c'}, 'a': {'c'}, 'b': {'c'}, 'e': set([])}
+
+    Testing border cases
+    >>> assert edge_reversed_graph(dict(e='', a='e')) == {'e': ['a'], 'a': []}
+    >>> assert edge_reversed_graph(dict(a='e', e='')) == {'e': ['a'], 'a': []}
+    """
+    # Pattern: Groupby logic
+
+    d = defaultdict(dst_nodes_factory)
+    for src, dst_nodes in g.items():
+        d.setdefault(src, dst_nodes_factory())  # add node if not present
+        for dst in dst_nodes:  # empty iterable does nothing
+            dst_nodes_append(d[dst], src)
+    return d
